from functools import partial
from multiprocessing import Pool
from typing import Any, Callable, Iterable, List, Optional


class Loader:
    def __call__(self, function: Callable, outputs: Iterable[Any]) -> Iterable[Any]:
        yield from map(function, outputs)


class BatchLoader(Loader):
    def __init__(self, batch_size: int, collate: Optional[Callable] = None) -> None:
        self._batch_size = batch_size
        self._collate = collate

    def collate(self, outputs: List[Any]) -> Any:
        if self._collate is None:
            return outputs
        return self._collate(outputs)

    def __call__(self, function: Callable, outputs: Iterable[Any]) -> Iterable[Any]:
        batch = []
        for raw_output in outputs:
            batch.append(raw_output)
            if len(batch) == self._batch_size:
                outputs = self.collate(batch)
                yield function(outputs)
                batch = []

        if batch:
            outputs = self.collate(batch)
            yield function(outputs)


class MultiprocessingLoader(Loader):
    @staticmethod
    def function(function: Callable, output: Any) -> List[Any]:
        return list(function(output))

    def __call__(self, function: Callable, outputs: Iterable[Any]) -> Iterable[Any]:
        with Pool() as pool:
            yield from pool.imap(partial(self.function, function), outputs)



import unittest
from typing import Iterable

from stepizer.step import Step


def read_paths(directory: str) -> Iterable[str]:
    for i in range(len(directory)):
        yield f'{directory}/image_{i}.jpg'


def read_image(path: str) -> list:
    return [ord(char) for char in path]


def detect_faces(image: list) -> Iterable[list]:
    for i in range(0, len(image), 7):
        yield image[i:i + 7]


def save(face: list, path: str = 'None') -> str:
    return path + f'.{sum(face)}.save'


class TestPipeline(unittest.TestCase):
    def setUp(self) -> None:
        self.pipeline = (
            Step(read_paths, is_generator=True, cache_mode='add')
            | read_image
            | Step(detect_faces, is_generator=True)
            | Step(save, args_mapping=('', 'read_paths'))
        )

    def test_pipeline(self) -> None:
        expected = [
            'dir/image_0.jpg.677.save',
            'dir/image_0.jpg.611.save',
            'dir/image_0.jpg.103.save',
            'dir/image_1.jpg.677.save',
            'dir/image_1.jpg.612.save',
            'dir/image_1.jpg.103.save',
            'dir/image_2.jpg.677.save',
            'dir/image_2.jpg.613.save',
            'dir/image_2.jpg.103.save',
        ]

        actual = self.pipeline.run('dir')

        self.assertListEqual(expected, actual)



from pathlib import Path

from setuptools import find_packages, setup

from stepizer import __version__

setup(
    name='stepizer',
    version=__version__,
    author='Mateusz Baran',
    author_email='mateusz.baran.sanok@gmail.com',
    maintainer='Mateusz Baran',
    maintainer_email='mateusz.baran.sanok@gmail.com',
    license='MIT',
    url='https://gitlab.com/mateusz.baran/stepizer',
    description='The tool for better organization of the workflow.',
    long_description=Path('README.md').read_text(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    python_requires='>=3.8, <4',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
    ],
    extras_require={
        'dev': [
            'bump2version>=1.0.1',
            'pytest>=6.1.2',
            'pytest-cov>=2.12.1',
            'pytest-flake8>=1.0.6',
            'pytest-isort>=2.0.0',
            'pytest-mypy>=0.8.1',
            'pytest-pycodestyle>=2.2.0',
        ],
    },

)



from __future__ import annotations

from functools import partial
from typing import Any, Callable, Dict, Iterable, Optional, Tuple

from stepizer.loader import Loader

ArgsType = Tuple[Any, ...]
KwargsType = Dict[str, Any]


class Step:
    def __init__(
        self,
        /,
        callable_: Callable,
        *,
        args: Optional[ArgsType] = None,
        kwargs: Optional[KwargsType] = None,
        name: Optional[str] = None,
        loader: Optional[Loader] = None,
        args_mapping: Optional[ArgsType] = None,
        kwargs_mapping: Optional[KwargsType] = None,
        is_generator: bool = False,
        cache_mode: str = 'ignore',
    ) -> None:
        self._callable = callable_

        self._args = args or tuple()
        self._kwargs = kwargs or dict()
        self._name = name or self._callable.__name__
        self._loader = loader or Loader()
        self._args_mapping = args_mapping
        self._kwargs_mapping = kwargs_mapping

        self._is_generator = is_generator
        self._cache_mode = cache_mode

        self._next_step: Optional[Step] = None

        if self._is_generator:
            self._call = self._generator
        else:
            self._call = self._function

    @property
    def callable(self) -> Callable:
        return self._callable

    @property
    def args(self) -> ArgsType:
        return self._args

    @property
    def kwargs(self) -> KwargsType:
        return self._kwargs

    @property
    def name(self) -> str:
        return self._name

    @property
    def loader(self) -> Loader:
        return self._loader

    @property
    def args_mapping(self) -> Optional[ArgsType]:
        return self._args_mapping

    @property
    def kwargs_mapping(self) -> Optional[KwargsType]:
        return self._kwargs_mapping

    @property
    def is_generator(self) -> bool:
        return self._is_generator

    @property
    def cache_mode(self) -> str:
        return self._cache_mode

    @property
    def next_step(self) -> Optional[Step]:
        return self._next_step

    def iter_steps(self) -> Iterable[Step]:
        yield self
        if self._next_step:
            yield from self._next_step.iter_steps()

    @classmethod
    def wrap(cls, callable_: Callable) -> Step:
        if isinstance(callable_, cls) and callable_._next_step is None:
            return callable_
        return cls(callable_)

    @classmethod
    def chain(cls, callable_: Callable, *callables: Callable) -> Step:
        step = cls.wrap(callable_)
        for c in callables:
            step.link(c)
        return step

    def link(self, callable_: Callable) -> Step:
        if self._next_step is None:
            self._next_step = Step.wrap(callable_)
        else:
            self._next_step.link(callable_)
        return self

    __or__ = link

    def __call__(self, *args, _cache: Optional[KwargsType] = None, **kwargs) -> Iterable[Any]:
        for outputs in self._loader(
            function=partial(self._generate_output, cache=_cache or dict()),
            outputs=self._call(*self._args, *args, **self._kwargs, **kwargs),
        ):
            yield from outputs

    def run(self, *args, **kwargs) -> Any:
        output = self(*args, **kwargs)
        if any(step._is_generator for step in self.iter_steps()):
            return list(output)
        return next(iter(output))

    def execute(self, *args, **kwargs) -> None:
        for _ in self(*args, **kwargs):
            pass

    def _map_arguments(self, output: Any, cache: KwargsType) -> Tuple[ArgsType, KwargsType]:
        if self._args_mapping is None and self._kwargs_mapping is None:
            return (output,), dict()

        args = list()
        if self._args_mapping is not None:
            for name in self._args_mapping:
                args.append(cache[name] if name else output)

        kwargs = dict()
        if self._kwargs_mapping is not None:
            for argument, name in self._kwargs_mapping.items():
                kwargs[argument] = cache[name] if name else output

        return tuple(args), kwargs

    def _update_cache(self, output: Any, cache: KwargsType) -> KwargsType:
        if self._cache_mode == 'ignore':
            return cache.copy()
        if self._cache_mode == 'add':
            return {**cache, self._name: output}
        if self._cache_mode == 'update':
            return {**cache, **output}
        raise ValueError(f"Cache mode {self._cache_mode} is not supported")

    def _generate_output(self, output: Any, cache: KwargsType) -> Iterable[Any]:
        if self._next_step is None:
            yield output
        else:
            cache = self._update_cache(output=output, cache=cache)
            args, kwargs = self._next_step._map_arguments(output=output, cache=cache)
            yield from self._next_step(*args, _cache=cache, **kwargs)

    def _function(self, *args, **kwargs) -> Iterable[Any]:
        yield self._callable(*args, **kwargs)

    def _generator(self, *args, **kwargs) -> Iterable[Any]:
        yield from self._callable(*args, **kwargs)



import unittest
from unittest.mock import Mock

from stepizer.loader import Loader
from stepizer.step import Step


class TestStepProperties(unittest.TestCase):
    def setUp(self) -> None:
        self.callable = Mock(__name__='Mock')
        self.step = Step(self.callable)

    def test_callable(self) -> None:
        function = self.step.callable

        self.assertIs(function, self.callable)

    def test_args(self) -> None:
        args = self.step.args

        self.assertIsInstance(args, tuple)

    def test_kwargs(self) -> None:
        kwargs = self.step.kwargs

        self.assertIsInstance(kwargs, dict)

    def test_name(self) -> None:
        name = self.step.name

        self.assertIsInstance(name, str)
        self.assertEqual(self.callable.__name__, name)

    def test_loader(self) -> None:
        loader = self.step.loader

        self.assertIsInstance(loader, Loader)

    def test_args_mapping(self) -> None:
        args_mapping = self.step.args_mapping

        self.assertIsNone(args_mapping)

    def test_kwargs_mapping(self) -> None:
        kwargs_mapping = self.step.kwargs_mapping

        self.assertIsNone(kwargs_mapping)

    def test_is_generator(self) -> None:
        generator = self.step.is_generator

        self.assertIsInstance(generator, bool)
        self.assertFalse(generator)

    def test_cache_mode(self) -> None:
        cache_mode = self.step.cache_mode

        self.assertIsInstance(cache_mode, str)
        self.assertEqual('ignore', cache_mode)

    def test_next_step(self) -> None:
        next_step = self.step.next_step

        self.assertIsNone(next_step)
